/*
 * Copyright (C) 2005 Cesar Kallas
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * E-mail: cesarkallas@gmx.net
 * http://cesarkallas.soulivre.org
 *
 */

/**
* Funcoes do servidor
* @author Cesar Kallas - cesarkallas@gmx.net
* @since 23/11/2005
*/
package estella.servidor;
import  estella.lib.*;

import  java.io.IOException;
import  java.lang.Thread;
import  java.net.Socket;
import  java.net.SocketTimeoutException;
import  java.util.StringTokenizer;

public class Servidor implements Runnable
{
  private ConexaoServidor conexaoServidor;
  private Thread t_conexao;
  private Log LOG;

  public Servidor()
  {
    LOG = new Log("Servidor");
		try {
			conexaoServidor = new ConexaoServidor();
		}
		catch(Exception FalhaServidor) {
			LOG.falha(FalhaServidor.getMessage());
			return;
		}
  }
  
  /**
    * Inicia um novo servico de escuta em uma porta
    */
  public void run()
  {
    Socket sck_cliente;
    conexaoServidor.inicia();
    LOG.info("Iniciado @ " + conexaoServidor.porta());
		
		// Seta IP e PORTA no qual o servidor foi instanciado
		Dados.ip = conexaoServidor.status().getInetAddress().getHostAddress();
		Dados.porta = conexaoServidor.status().getLocalPort();

    // Escuta eternamente requisicao no servidor
    while(true) {
      try {
        sck_cliente = conexaoServidor.escuta();
      }
      catch(Exception FalhaEscutarServidor)
      {
        LOG.falha(FalhaEscutarServidor.getMessage());
        return;
      }
      
      // Nova requisicao, ela eh null ? (pode ser quando o cliente conecta e desconecta)
      // Se ok, cria uma nova thread para tratar a requisicao
      if(sck_cliente != null) 
        try {
					LOG.info("Novo cliente: " + sck_cliente.getInetAddress().getHostAddress() + ":" + sck_cliente.getPort() + ":" + conexaoServidor.porta());
          t_conexao = new Thread(new Cliente(sck_cliente));
          t_conexao.start();
        }
        catch(Exception threadConexao) {
          LOG.falha(threadConexao.getMessage());
          return;
        }
    }
  }
}

/**
* Thread para tratar requisicao recebida pelo servidor
* @param sck_cliente Socket com a conexao do cliente
* @since 23/11/2005
* @author Cesar Kallas - cesarkallas at gmx . net
*/
class Cliente implements Runnable
{
  private Socket sck_cliente;
  private Conexao conexao;
  private Log LOG;
  private Constantes CON;
	private Descritor descritor;
	private String mensagem;
	private Dados dados;
	private Thread t_trataDescritor;
		
  public Cliente(Socket sck_cliente)
  {
    LOG = new Log("Servidor");
    CON = new Constantes();
    dados = new Dados();
		conexao = new Conexao(sck_cliente);
  }

  public void run()
  {
		// Trata pedido de conexao do cliente
		try {
			conexao.inicia();
			conexao.setTimeout(CON.SOCKET_SERVER_TIMEOUT);
			mensagem = conexao.recebe();
			if(mensagem.indexOf(CON.GNUTELLA) == -1)
				LOG.falha("Cabecalho de autenticacao invalido - " + conexao.status());
			else
				if(conexao.envia(CON.GNUTELLA_OK + '\n')) {
					conexao.setTimeout(0);
					synchronized(Dados.onlineClientes) {
						Dados.onlineClientes.adiciona(conexao);
					}
					trataCliente();
				}
			else
				LOG.falha("ao enviar resposta de pedido de conexao para " + conexao.status());
		}
		catch(Exception FalhaAoConectar)
		{}

		try {
			conexao.fecha();
		}
		catch(Exception FalhaDesconectarCliente)
		{}
		return;
  }
		
	public void trataCliente()
	{
		// Cliente conectado, tratar descritores do cliente
		while(conexao.status() != null) {
			try {
				mensagem = new String(conexao.recebe());
			}
			catch(Exception ReceberDescritor)
			{
				LOG.falha(ReceberDescritor.getMessage());
				return;
			}
			descritor = new Descritor(mensagem);
			t_trataDescritor = new Thread(new trataDescritor(descritor, conexao));
			t_trataDescritor.start();
		}
		LOG.info("Cliente desconectado");
		return;
	}
}

class trataDescritor implements Runnable
{
	private Descritor descritor, descritorResposta;
	private Log LOG;
	private Constantes CON;
	private Conexao conexao;
	private Payload payload;
		
	public trataDescritor(Descritor descritor, Conexao conexao)
	{
		this.descritor = descritor;
		this.conexao = conexao;
		this.CON = new Constantes();
		this.LOG = new Log("Servidor_trataDescritor");
	}
	
	public void run()
	{
		// Tamanho minimo de descritor 20 (id + ttl + hops + payload length)
		if(descritor.retorna().length < 20)
			return;
		
		if(!adicionaDescritor())
			return;
		
		switch(descritor.tipo())
		{
			case (byte)00000000: { ping(); break; }
			case (byte)00000001: { pong(); break; }
			case (byte)10000000: { query(); break; }
			case (byte)10000001: { queryhits(); break; }
			default: return;
		}
	}
	
	public boolean adicionaDescritor()
	{
		try {
			// Adiciona na lista de descritores
			synchronized(Dados.listaDescritores) {
				if(!Dados.listaDescritores.adiciona(descritor)) {
					LOG.falha("ao adicionar descritor na lista de descritores");
					return false;
				}
				// Incrementa hops e decrementa ttl
				descritor.ttl((int)(0xFF & descritor.ttl())-1);
				descritor.hops((int)(0xFF & descritor.hops())+1);
			}
			return true;
		}
		catch(Exception E) {
			LOG.falha(E.getMessage());
			return false;
		}
	}
	
	/**
	* Responde e repassa o pong de um cliente
	*/
	private void ping()
	{
		LOG.info("ping id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") -> "  + conexao.status());
		
		// Prepara pong local
		payload = new Payload();
		
		descritorResposta = new Descritor(descritor.id(), CON.PONG, descritor.ttl(), descritor.hops(), 
			payload.pong(Dados.porta, Dados.ip, Dados.numArqComp, Dados.tamArqComp));

		// Tenta responder o pong 
		if(conexao.envia(new String(descritorResposta.retorna()) + '\n')) {
			if(repassaDescritor())
				LOG.info("pong id(" + descritorResposta.stringId() + ") ttl(" + descritorResposta.ttl() + ") hops(" + descritorResposta.hops() +  ") -> "  + conexao.status());
		} // Nao consegui enviar PONG local (cliente desconectou ?) ~ removo o cliente da lista
		else
			LOG.falha("enviar PONG para " + conexao.status());
	}
	
	/*
	* Recebe e responde ao ping do cliente
	*/
	private void pong()
	{
		LOG.info("pong de " + conexao.status());
	}
	
	private void query()
	{		
		StringTokenizer resultadoProcura;
		
		LOG.info("query id(" + descritor.stringId() + ") ttl(" + descritor.ttl() + ") hops(" + descritor.hops() +  ") -> "  + conexao.status());
		Payload query = new Payload(CON.QUERY, descritor.payload());
		LOG.info("velocidade minima: " + query.queryMinVelocidade());
		synchronized(Dados.arquivos) {
			resultadoProcura = new StringTokenizer(Dados.arquivos.existe(query.queryTermo()), CON.SEPARADOR);
		}
		Payload payload = new Payload();
		
		byte[] hit = new byte[1];
		byte[] hit2 = new byte[1];
		byte[] hits = new byte[1];
		int numHits=0;
		
		if(resultadoProcura.hasMoreTokens())
			hits = payload.hit(Integer.parseInt(resultadoProcura.nextToken()),
				Integer.parseInt(resultadoProcura.nextToken()), resultadoProcura.nextToken());
		System.out.println("..." + hits.length);
		
		while(resultadoProcura.hasMoreTokens() && numHits < 125) {
			hits = payload.concatenaHits(hits, payload.hit(Integer.parseInt(resultadoProcura.nextToken()),
				Integer.parseInt(resultadoProcura.nextToken()), resultadoProcura.nextToken()));
			numHits++;
			System.out.println("..." + hits.length);
		}
		
		System.out.println("..." + hits.length);
		
		byte[] resultset = payload.queryhits(numHits, Dados.porta, Dados.ip, hits);
		descritorResposta = new Descritor(descritor.id(), CON.QUERYHITS, descritor.ttl(), descritor.hops(), resultset);

		// Tenta responder o queryhits
		if(conexao.envia(new String(descritorResposta.retorna()) + '\n')) {
			if(repassaDescritor())
				LOG.info("queryhits id(" + descritorResposta.stringId() + ") ttl(" + descritorResposta.ttl() + ") hops(" + descritorResposta.hops() +  ") -> "  + conexao.status());
		} // Nao consegui enviar PONG local (cliente desconectou ?) ~ removo o cliente da lista
		else
			LOG.falha("enviando queryhits a " + conexao.status());
		System.out.println("tamanho do queryhits no serivdor:" + descritorResposta.payload().length + " num=" + numHits);
	}
	
	private void queryhits()
	{
		LOG.info("queryhits de " + conexao.status());
	}
	
	public boolean repassaDescritor()
	{
		try {
			// Verifica se deve repassar para servents que ele ta conectado o descritor recebido
			if(descritor.ttl() > 0)
				// Deve repassar os pongs, estou conectado em alguem ? se sim, envia
				if(Dados.onlineServents.tamanho() > 0)
					synchronized(Dados.onlineServents) {
						// Envia para todos servents que estou conectado
						Dados.onlineServents.envia(new String(descritor.retorna())+'\n');
						return true;
					}
			return false;
		}
		catch(Exception FalhaRepassarDescritor) {
			LOG.falha("falha repassar descritor " + FalhaRepassarDescritor.getMessage());
			return false;
		}
	}
}
